#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <pthread.h> 
#include <stdio.h>
#include <time.h>

#define DEADLOCK 0
#define IMAGE_AMOUNT 25
#define WAITING 0
#define WORKING 1
#define FINISHED 2

pthread_t thr_1; 
pthread_t thr_2; 

int lines_index = 0;
int s; /* semaphore */
int s_writer_init = 5; /* semaphore initial value for writer thread */
int s_reader_init = 15; /* semaphore initial value for reader thread */
int LINES[IMAGE_AMOUNT];
char reader_status = WAITING;

pthread_mutex_t mtx_lock;

void thread_report(int thread) {

     if (!thread)
          printf("Writer: LINES: ");
     else
          printf("Reader: LINES: ");

     for (int i = 0; i < lines_index; i++)
          printf(".");

     printf("\n");

}

void *writer() { 

     while (reader_status != FINISHED) {
          if (reader_status != WORKING) {           
               pthread_mutex_lock(&mtx_lock); 
               s = s_writer_init;
               while (1) {
                    if (!DEADLOCK) 
                         /* Here, s acts as a semaphore. Every time we loop within the mutex, the semaphore
                          * decrements by 1. When s is zero, we break out of the loop, the lock is released,
                          * and so the reader thread has an opportunity to access LINES, etc. */ 
                         if (--s == 0) {
                              break;
                              reader_status = WORKING;
                         }
                    if (lines_index < IMAGE_AMOUNT) {
                         LINES[lines_index++] = rand() % 2055;
                         thread_report(0);
                    }
               }
               pthread_mutex_unlock(&mtx_lock); 
          }
     }

     printf("Writer done.\n");
     pthread_exit(NULL);

} 

void make_img(int line, int panel) {

     char cmd[250];

     sprintf(cmd,"bash ./addtext.sh %d %d", panel, line);
     printf("Reader: %s\n", cmd);         
     system(cmd);
           
}

void *reader() { 

     int img_id = 0;
     int images_generated = 0;

     while (images_generated < IMAGE_AMOUNT) {
          reader_status = WORKING;
          pthread_mutex_lock(&mtx_lock); 
          s = s_reader_init;
          while (1) {
               if (!DEADLOCK) 
                    if (--s == 0) {
                         reader_status = WAITING;
                         break;
                    }
               if (lines_index > 0 && images_generated++ < IMAGE_AMOUNT) {
                    thread_report(1);
                    make_img(LINES[--lines_index], img_id++);
               }
          }
          pthread_mutex_unlock(&mtx_lock); 
     }
     reader_status = FINISHED;
     printf("Reader done.\n");
     pthread_exit(NULL);

}

int main(void) { 

     int thread_return;
     int mutex_return;
     char cmd[250];

     /* Seed the random number generator with time. */
     srand(time(NULL));

	mutex_return = pthread_mutex_init(&mtx_lock, NULL);
     if (mutex_return) {
          printf("Mutex init error: %d\n", mutex_return);
		return(1); 
     }

     thread_return = pthread_create(&thr_1, NULL, writer, NULL);
     if (thread_return != 0) 
          return(thread_return);

     thread_return = pthread_create(&thr_2, NULL, reader, NULL);
     if (thread_return != 0) 
          return(thread_return); 

	pthread_join(thr_1, NULL); 
	pthread_join(thr_2, NULL); 

	pthread_mutex_destroy(&mtx_lock); 

     /* Combine panel images into stories: */ 
     for (int panel = 0; panel < IMAGE_AMOUNT; panel++) {
          if (panel > 2 && panel % 3 == 0) {
               sprintf(cmd,"convert /tmp/panel%d.png /tmp/panel%d.png /tmp/panel%d.png +append /tmp/story%d.png", 
                         panel-2, panel-1, panel, panel / 3); 
               printf("cmd: %s\n", cmd);         
               system(cmd);
          }
     }
     system("mkdir -p ./stories");
     system("cp /tmp/story*png ./stories");

     printf("Done.\n");

	return 0; 
} 

